#!/usr/bin/env python
# coding: utf-8

"""
 *
 * TurtleBot3RosOdomSub.py
 *
 *  Created on: 2021/11/09
 *      Author: Tsuyoshi Anazawa
 *
 *  Outline: Output odometory in csv format.
 *
"""
import os
from datetime import datetime, timedelta
from time import strftime
import pytz
import csv
import rospy
from nav_msgs.msg import Odometry
import rosparam
import tf
import rospy_tutorials
from geometry_msgs.msg import TransformStamped
from geometry_msgs.msg._PoseStamped import PoseStamped

filePath = ""
fileName = "/odom.csv"
robot_id = ""
robot_type = ""
space = ""
# def odomCallback(msg):
#     try:
#         global filePath, tf_prefix
#         hdr_flag = False
#         jst = pytz.timezone('Asia/Tokyo')
#         dt = datetime.now(jst)
#         iso_time = dt.strftime("%FT%T.%f")[:-3]
#         time_zone = dt.strftime("%z")
#
#         dst = iso_time + time_zone
#
#         if not os.path.isfile(filePath):
#             hdr = [ "robot_id", "datetime", "odom_pos_x", "odom_pos_y", "odom_pos_z", 
#                     "odom_ori_x", "odom_ori_y", "odom_ori_z", "odom_ori_w" ]
#             hdr_flag = True
#
#         if hdr_flag:
#             with open(filePath, 'a') as f:
#                 writer = csv.writer(f)
#                 writer.writerow(hdr)
#         else:
#             with open(filePath, 'a') as f:
#                 writer = csv.writer(f)
#                 writer.writerow( [ tf_prefix, dst, msg.pose.pose.position.x, msg.pose.pose.position.y, 
#                                    msg.pose.pose.position.z, msg.pose.pose.orientation.x,
#                                    msg.pose.pose.orientation.y, msg.pose.pose.orientation.z, msg.pose.pose.orientation.w ] )
#
#     except Exception as err:
#         print(err)

# def odom_proc():
#     global filePath, fileName, tf_prefix
#     filePath = os.getcwd()
#     filePath = filePath + fileName
#
#     tf_prefix = rospy.get_param("/turtlebot_01/robot_state_publisher/tf_prefix", None)
#     if tf_prefix is None:
#     	tf_prefix = rospy.get_param("/tb3_0/robot_state_publisher/tf_prefix", None)
#
#     rospy.init_node('odom_proc', anonymous=True)
#     rospy.loginfo('odom_proc node started.')
#
#     # Subscribe robot odometry values.
#     rospy.Subscriber("/" + str(tf_prefix) + "/odom", Odometry, odomCallback)
#     # Call the callback function repeatedly.
#     rospy.spin()

def writeCsv():
    try:
        global filePath, robot_id, robot_type, space
        slash = "/"
        parent_frame = "map"
        child_frame = slash + robot_id + slash + "base_link"
        jst = pytz.timezone('Asia/Tokyo')
        hdr_flag = True
        
        # rospy.loginfo('file exist: ' + str(os.path.isfile(filePath)))
        # rospy.loginfo('file path: ' + str(os.path.abspath(filePath)))
        if not os.path.isfile(filePath):
            hdr = [ "id", "type", "space", "time", "seq", "secs", "nsecs", "frame_id", 
                   "pos_x", "pos_y", "pos_z", "ori_x", "ori_y", "ori_z", "ori_w" ]
            
            with open(filePath, 'a') as f:
                writer = csv.writer(f)
                writer.writerow(hdr)
                # hdr_flag = True
        
        tf_listener_ = tf.TransformListener()
        rate = rospy.Rate(1.0)
        
        while not rospy.is_shutdown():
            try:
                if not tf_listener_.canTransform(parent_frame, child_frame, rospy.Time(0)):
                    continue
                tf_listener_.waitForTransform(parent_frame, child_frame, rospy.Time(0), rospy.Duration(15.0))
                time_stamp = rospy.Time.now()
                (trans, rot) = tf_listener_.lookupTransform(parent_frame, child_frame, rospy.Time(0))
                dt = datetime.now(jst)
                iso_time = dt.strftime("%FT%T.%f")[:-3]
                time_zone = dt.strftime("%z")
                dst = iso_time + time_zone
        
                # if hdr_flag:
                #     with open(filePath, 'a') as f:
                #         writer = csv.writer(f)
                #         writer.writerow(hdr)
                #         hdr_flag = False
                # else:
                #     with open(filePath, 'a') as f:
                #         writer = csv.writer(f)
                #         writer.writerow( [ tf_prefix, dst, trans[0], trans[1], '0.0', 
                #                           rot[0], rot[1], rot[2], rot[3] ] )
                rospy.loginfo(str(robot_id) + ' ' + str(robot_type) + ' ' + str(space) + ' ' + str(dst) + ' ' + str(0) + ' ' 
                              + str(time_stamp.secs) + ' ' + str(time_stamp.nsecs) + ' ' + parent_frame + ' ' + str(trans[0]) + ' ' 
                              + str(trans[1]) + ' ' + str(trans[2]) + ' ' + str(rot[0]) + ' ' + str(rot[1]) + ' ' 
                              + str(rot[2]) + ' ' + str(rot[3]))
                with open(filePath, 'a') as f:
                    writer = csv.writer(f)
                    writer.writerow( [ robot_id, robot_type, space, dst, 0, time_stamp.secs, 
                                      time_stamp.nsecs, parent_frame, trans[0], trans[1], 0.0, 
                                      rot[0], rot[1], rot[2], rot[3] ] )
                        
            except (tf.LookupException, tf.ConnectivityException, tf.ExtrapolationException):
                continue
            finally:
                rate.sleep()
    except Exception as err:
        print(err)

def tf_proc():
    global filePath, fileName, robot_id, robot_type, space
    filePath = os.getcwd()
    filePath = filePath + fileName
    
    robot_id = rospy.get_param("id", None)
    robot_type = rospy.get_param("type", None)
    space = rospy.get_param("space", None)
    
    rospy.init_node('tf_proc', anonymous=True)
    rospy.loginfo('tf_proc node started.')
    
    writeCsv()

if __name__ == '__main__':
    try:
        # odom_proc()
        tf_proc()
        
    except rospy.ROSInterruptException:
        pass
    
    finally:
        # rospy.loginfo('odom_proc node finished.')
        rospy.loginfo('tf_proc node finished.')
