/*
 * TurtleBot3RosOdomSub.cpp
 *
 *  Created on: 2021/11/09
 *      Author: Tsuyoshi Anazawa
 *
 *  Outline: This controller for outputting odometry in CSV format.
 */

#include <cnoid/SimpleController>
#include <ros/node_handle.h>
#include <ros/ros.h>
#include <nav_msgs/Odometry.h>
#include <std_msgs/String.h>
#include <iostream>
#include <fstream>
#include <tf/transform_broadcaster.h>
#include <tf/transform_listener.h>
#include "delivery_robot/r_navi_command.h"

using namespace std;
using namespace cnoid;

class TurtleBot3RosOdomSub : public SimpleController
{
	ros::NodeHandle nh_;
	ros::Subscriber odom_sub_;
	ros::Subscriber simtime_sub_;
	ros::Subscriber r_navi_cmd_sub_;
	tf::TransformListener tf_listener_;

//	string tf_prefix;
	string robot_id;
	string robot_type;
	string space;
	FILE* fp;
	static const int BUF_SIZE = 100;
	char str[BUF_SIZE];
	string filePath;
	string fileName = "odom.csv";
	string cmd = "pwd | tr '\n' '/'";
	const string COMMA = ",";
	ofstream ofs;
	string simtime;

	bool check_recieve = true;
	bool get_odom_flg = false;

	SimpleControllerIO* io;
	double update_period;
	double update_rate = 1.0;
	double lastTime;

	const string PARENT_FRAME = "map";
	const string CHILD_FRAME = "base_link";
	const string SLASH = "/";

public:
	virtual bool configure(SimpleControllerConfig* config) override
	{
		return true;
	}

	virtual bool initialize(SimpleControllerIO* io) override
	{
		this->io = io;
		nh_.getParam("id", robot_id);
//		nh_.getParam(SLASH + robot_id + "/robot_state_publisher/tf_prefix", tf_prefix);
		nh_.getParam("type", robot_type);
		nh_.getParam("space", space);

//		odom_sub_ = nh_.subscribe(tf_prefix + "/odom", 1, &TurtleBot3RosOdomSub::odomCallback, this);
		simtime_sub_ = nh_.subscribe("/simtime", 1, &TurtleBot3RosOdomSub::simtimeCallback, this);
		r_navi_cmd_sub_ = nh_.subscribe("/robot_bridge/" + robot_id + "/navi_cmd", 1, &TurtleBot3RosOdomSub::r_navi_cmdCallback, this);

		if((fp = popen(cmd.c_str(), "r")) != NULL){
			// Open a process and execute the command.
			while(fgets(str, sizeof(str), fp) != NULL){
				// Read command results line by line.
				// Get current directory.
				filePath += str;
			}
			// Close the process.
			pclose(fp);
		}

		filePath = filePath + fileName;

//		io->os() << filePath << endl;

		ifstream ifs(filePath);
		bool hdr_flag = false;
		string hdr;
		if(ifs.fail()){
			hdr = "id" + COMMA +
				  "type" + COMMA +
				  "space" + COMMA +
				  "time" + COMMA +
//				  "simtime" + COMMA +
				  "seq" + COMMA +
				  "secs" + COMMA +
				  "nsecs" + COMMA +
				  "frame_id" + COMMA +
				  "pos_x" + COMMA +
				  "pos_y" + COMMA +
				  "pos_z" + COMMA +
				  "ori_x" + COMMA +
				  "ori_y" + COMMA +
				  "ori_z" + COMMA +
				  "ori_w";
			hdr_flag = true;
		}

		ofs.open(filePath, ios::app);

		if(hdr_flag){
			ofs << hdr << endl;
		}

		update_period = 1.0 / update_rate;
		lastTime = io->currentTime();

		return true;
	}

	void odomCallback(const nav_msgs::Odometry& msg)
	{
//		tf::Quaternion quat(msg.pose.pose.orientation.x, msg.pose.pose.orientation.y, msg.pose.pose.orientation.z, msg.pose.pose.orientation.w);
//		double r, p, y;
//		tf::Matrix3x3(quat).getRPY(r, p, y);
//
//		ofs << tf_prefix << COMMA << iso8601ex() << COMMA << msg.pose.pose.position.x << COMMA
//				<< msg.pose.pose.position.y << COMMA << msg.pose.pose.position.z << COMMA
//				<< r << COMMA << p << COMMA
//				<< y << COMMA << endl;
		if(get_odom_flg){
//			ofs << tf_prefix << COMMA << iso8601ex() << COMMA << simtime << COMMA
//					<< msg.pose.pose.position.x << COMMA << msg.pose.pose.position.y << COMMA
//					<< msg.pose.pose.position.z << COMMA << msg.pose.pose.orientation.x << COMMA
//					<< msg.pose.pose.orientation.y << COMMA << msg.pose.pose.orientation.z << COMMA
//					<< msg.pose.pose.orientation.w << endl;
			ofs << robot_id << COMMA << iso8601ex() << COMMA << msg.pose.pose.position.x << COMMA
					<< msg.pose.pose.position.y << COMMA << msg.pose.pose.position.z << COMMA
					<< msg.pose.pose.orientation.x << COMMA << msg.pose.pose.orientation.y << COMMA
					<< msg.pose.pose.orientation.z << COMMA << msg.pose.pose.orientation.w << endl;
		}
	}

	void simtimeCallback(const std_msgs::String& msg)
	{
		simtime = msg.data.c_str();

		double currentTime = io->currentTime();
		double updateSince = currentTime - lastTime;
		if(updateSince > update_period){
			tf::StampedTransform stamped_tf_;
			try {
				tf_listener_.waitForTransform(PARENT_FRAME, SLASH + robot_id + SLASH + CHILD_FRAME, ros::Time(0), ros::Duration(10));
				tf_listener_.lookupTransform(PARENT_FRAME, SLASH + robot_id + SLASH + CHILD_FRAME, ros::Time(0), stamped_tf_);

				if(get_odom_flg){
//					ofs << tf_prefix << COMMA << iso8601ex() << COMMA << simtime << COMMA
//							<< stamped_tf_.getOrigin().x() << COMMA << stamped_tf_.getOrigin().y() << COMMA
//							<< stamped_tf_.getOrigin().z() << COMMA << stamped_tf_.getRotation().x() << COMMA
//							<< stamped_tf_.getRotation().y() << COMMA << stamped_tf_.getRotation().z() << COMMA
//							<< stamped_tf_.getRotation().w() << endl;
					ofs << robot_id << COMMA << robot_type << COMMA << space << COMMA << iso8601ex() << COMMA
							<< 0.0 << COMMA << stamped_tf_.stamp_.sec << COMMA << stamped_tf_.stamp_.nsec << COMMA
							<< stamped_tf_.frame_id_ << COMMA << stamped_tf_.getOrigin().x() << COMMA
							<< stamped_tf_.getOrigin().y() << COMMA << 0.0 << COMMA
							<< stamped_tf_.getRotation().x() << COMMA << stamped_tf_.getRotation().y() << COMMA
							<< stamped_tf_.getRotation().z() << COMMA << stamped_tf_.getRotation().w() << endl;
				}
				lastTime += update_period;
			} catch(tf::TransformException& ex){
				ROS_ERROR("%s", ex.what());
				lastTime += update_period;
			}
		}
	}

	void r_navi_cmdCallback(const delivery_robot::r_navi_command& msg)
	{
		if(check_recieve){
			get_odom_flg = true;
			check_recieve = false;
		}
	}

	string iso8601ex()
	{
		char iso_time[30];
		char time_zone[10];
		char dst[50];
		struct timeval ct;
		struct tm* local_time;

		memset(iso_time, 0, sizeof(iso_time));
		memset(time_zone, 0, sizeof(time_zone));
		memset(dst, 0, sizeof(dst));

		gettimeofday(&ct, NULL);
		local_time = localtime(&ct.tv_sec);

		strftime(iso_time, sizeof(iso_time) - 1, "%FT%T", local_time);
		strftime(time_zone, sizeof(time_zone) - 1, "%z", local_time);

		sprintf(dst, "%s.%03lu%s", iso_time, (ct.tv_usec + 500) / 1000, time_zone);

		string time_str = dst;

		return time_str;
	}

	virtual bool control() override
	{
		return true;
	}

	virtual void stop() override
	{
		ofs.close();
		odom_sub_.shutdown();
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot3RosOdomSub)
