#include <cnoid/SimpleController>
#include <cnoid/EigenUtil>
#include <cnoid/Link>
#include <vector>
#include <string>
#include <math.h>

//ROS
#include <ros/node_handle.h>
#include <std_msgs/Float32.h>
#include <geometry_msgs/PoseArray.h>
#include <geometry_msgs/Pose.h>
#include <tf/transform_broadcaster.h>
#include <tf/transform_broadcaster.h>
#include <Eigen/Geometry>

using namespace std;

using namespace cnoid;



class RailController : public cnoid::SimpleController
{
    Link* joint_CameraBase;
    Link* joint_CameraNeck;
    Link* joint_Camera;
    double f;         //joint_CameraBase value 
    double move_speed; //move speed m/s
    bool move_flag;    //move flag
    double limite;    // joint_CameraBase max value
    double cameraT[3];
    double camerabase_value;
    double preposition;
    double toposition;
    SimpleControllerIO* io;
    //ROS topic
    ros::NodeHandle node;
    ros::Subscriber robomove_sub_;
    ros::Publisher roboposition_pub_;

public:

    virtual bool configure(SimpleControllerConfig* config) override
    {
        robomove_sub_ = node.subscribe("cameramove", 1, &RailController::robotmoveCallback,this);
        roboposition_pub_ = node.advertise<geometry_msgs::Pose>("cameraposition", 10);
    
        return true;
    }
    
    virtual bool initialize(SimpleControllerIO* io) override
    {
        cout<<"initialize"<<endl;
        this->io = io;
        //CameraBase Joint Setup
        joint_CameraBase = io->body()->link("CameraBase");
        joint_CameraBase->setActuationMode(Link::LinkPosition | Link::JOINT_FORCE);
        io->enableInput(joint_CameraBase, JOINT_DISPLACEMENT | JOINT_VELOCITY | LinkPosition);
        joint_CameraBase->setActuationMode(Link::JOINT_VELOCITY);
        io->enableOutput(joint_CameraBase);
        //CameraNeck Joint Setup
        joint_CameraNeck = io->body()->link("CameraNeck");
        joint_CameraNeck->setActuationMode(Link::LinkPosition);
        io->enableInput(joint_CameraNeck, JOINT_DISPLACEMENT | JOINT_VELOCITY | LinkPosition);
        joint_CameraNeck->setActuationMode(Link::JointDisplacement);
        io->enableOutput(joint_CameraNeck);
        //Camerabody Joint Setup
        joint_Camera = io->body()->link("RangeCamera");
        joint_Camera->setActuationMode(Link::LinkPosition);
        io->enableInput(joint_Camera,  JOINT_VELOCITY | LinkPosition);
        f=0;
        limite=1.0;
        move_speed = 0.1;
        move_flag=true;
        preposition = 0;
        toposition = 0;
        return true;
    }

    virtual bool control() override
    {

        // Rail move
        if(move_flag == false && joint_CameraBase->dq_target()==0){
            check_moveposition();
            setmove_speed();
            move_flag= true;

        }
        //J̈ړAJ̈ʒuɒBꍇAړ̖ړIʒuɂꍇ~
        if(check_purpose_position() || check_limiteover()){
             joint_CameraBase->dq_target() = 0;
             //cout<<"joint_CameraBase->dq_target()="<<joint_CameraBase->dq_target()<<endl;
             //move_flag=false;
        }
        //̕
        joint_CameraNeck->dq_target()=0;
        //J̈ʒuo
        push_cameraposition();
        return true;

    }

    void check_moveposition()
    {
        //ړ\̈ʒu([)
        toposition = joint_CameraBase->q() + f ;
        //ړÖʒu([)
        preposition = joint_CameraBase->q();
        if(toposition < 0){
            toposition = 0;
        }else if(toposition > limite){
            toposition = limite;
        }else{
            toposition = joint_CameraBase->q() + f;
        }
        cout<<"check_preposition toposition ="<<toposition<<endl;
        cout<<"check_preposition preposition ="<<joint_CameraBase->q()<<endl;
        
        f=0;
        //return true;
    }
    void setmove_speed()
    {
        if(toposition > preposition ) joint_CameraBase->dq_target() = move_speed;
        else if(toposition < preposition ) joint_CameraBase->dq_target() = (-1)*move_speed;
        else joint_CameraBase->dq_target() = 0;
        cout<<"setmove_speed move speed ="<<joint_CameraBase->dq_target()<<endl;
        //return true;
    }
    bool check_purpose_position()
    {
        if(toposition > preposition ){
            //cout<<"check_purpose_position1  q ="<< joint_CameraBase->q()<<endl;
            //cout<<"check_purpose_position1 tq ="<< toposition<<endl;
            if (joint_CameraBase->q() >= toposition ) return true;
        }
        else{
            //cout<<"check_purpose_position2  q ="<< joint_CameraBase->q()<<endl;
            ///cout<<"check_purpose_position2 tq ="<< toposition<<endl;
            if (joint_CameraBase->q() <= toposition ) return true;
        }
        return false;
    }
    bool check_limiteover(){
           //ʒui[j0菬 @0ȉɈړꍇ
        if (joint_CameraBase->q() < 0 && joint_CameraBase->dq_target() < 0){
             joint_CameraBase->q() = 0;
             return true;
        }
           //ʒui[jEl傫 @EȏɈړꍇ
        if ( limite < joint_CameraBase->q() && joint_CameraBase->dq_target() > 0){
             joint_CameraBase->q() = limite;
             return true;
        }


        return false;
    }
    void robotmoveCallback(const std_msgs::Float32& input)
    {
        cout<<"robotmoveCallback"<<endl;
        f=input.data;
        cout<<"robotmove value:"<<input.data<<endl;
        move_flag=false;
    }
    geometry_msgs::Point getCamerapositionXYZ()
    {
        auto p = joint_Camera->translation() ;
        auto T = joint_Camera->position();
        double z = T.translation().z();
        auto cameraR = T.rotation();
        //Camera Position
        geometry_msgs::Point geometry_point;
        geometry_point.x= p.x();
        geometry_point.y= p.y();
        geometry_point.z= z;

        return geometry_point;
    }
    geometry_msgs::Quaternion  getCamerapositionquat()
    {
        auto T = joint_Camera->position();
        Vector3 rpy = rpyFromRot(T.rotation());
        tf::Quaternion quat=tf::createQuaternionFromRPY(rpy[0],rpy[1],rpy[2]);
        geometry_msgs::Quaternion geometry_quat;
        quaternionTFToMsg(quat, geometry_quat);
        return geometry_quat;        
    }
    void push_cameraposition()
    {
        // J̈ʒu擾
        geometry_msgs::Pose camera_pose;
        camera_pose.position = getCamerapositionXYZ();
        camera_pose.orientation = getCamerapositionquat();
        //J̈ʒuo
        roboposition_pub_.publish(camera_pose);

    }
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(RailController)

